from collections import defaultdict
from enum import Enum
from typing import Dict, Iterable, Optional, Set, Type
from pathlib import Path
from regex import escape, fullmatch, match, split, sub, IGNORECASE

from xlsxwriter import Workbook

from .config import CONFIG as _CONFIG
from .entity import Entity, Place, Person, Regest


class EntityType(Enum):
    PERSON = (Person, 'O')
    PLACE = (Place, 'M')

    @classmethod
    def from_entity_class(cls, entity_class: Optional[Type[Entity]]) -> Optional['EntityType']:
        if entity_class is None:
            return None
        for value in cls:
            if value.to_entity_class() == entity_class:
                return value
        raise ValueError('Unexpected entity class: {}'.format(entity_class))

    @classmethod
    def from_symbol(cls, symbol: Optional[str]) -> Optional['EntityType']:
        if symbol is None:
            return None
        for value in cls:
            if str(value) == symbol:
                return value
        raise ValueError('Unexpected symbol: {}'.format(symbol))

    def __str__(self) -> str:
        return self.value[1]

    def to_entity_class(self) -> Type[Entity]:
        return self.value[0]

    @classmethod
    def list_symbols(cls) -> str:
        return ', '.join(str(value) for value in cls)


class Approval(Enum):
    APPROVED = 'A'
    DISAPPROVED = 'N'
    TOO_LONG = '>'
    TOO_SHORT = '<'

    @classmethod
    def from_symbol(cls, symbol: Optional[str]) -> Optional['Approval']:
        if symbol is None:
            return None
        for value in cls:
            if str(value) == symbol:
                return value
        raise ValueError('Unexpected symbol: {}'.format(symbol))

    def __str__(self) -> str:
        return self.value

    @classmethod
    def list_symbols(cls) -> str:
        return ', '.join(str(value) for value in cls)


class Patch:
    def __init__(self, entity: Entity, canonical_form: Optional[str],
                 approval: Optional[Approval], note: Optional[str]):
        self.entity = entity
        self.canonical_form = canonical_form
        self.approval = approval
        self.note = note


CANONICAL_PLACES = {line.rstrip('\r\n')
                    for line
                    in open('/nlp/projekty/ahisto/public_html/named-entity-search/results/patchset-template-places-edited.txt', 'rt')}
CANONICAL_PERSONS = {line.rstrip('\r\n')
                     for line
                     in open('/nlp/projekty/ahisto/public_html/named-entity-search/results/patchset-template-persons-edited.txt', 'rt')}
assert not (CANONICAL_PLACES & CANONICAL_PERSONS)


approved = set()


class Patchset:
    CONFIG = _CONFIG['patch.Patchset']
    ROOT_PATH = Path(CONFIG['root_path'])
    MINIMUM_LENGTH = CONFIG.getint('minimum_length')
    MAXIMUM_LENGTH = CONFIG.getint('maximum_length')

    @classmethod
    def create_patchset_template(cls, filename: str, entities: Iterable[Entity],
                                 entities_to_regests: Dict[Entity, Set[Regest]],
                                 entity_type: str) -> None:

        entities = sorted(entities)

        # Create a plaintext file for power users
        with (cls.ROOT_PATH / filename).with_suffix('.txt').open('wt') as f:
            for entity in entities:
                print(entity, file=f)

        # Create a spreadsheet for standard users
        with Workbook((cls.ROOT_PATH / filename).with_suffix('.xlsx')) as workbook:

            # Declare formats
            default_format = workbook.add_format()

            bold_format = workbook.add_format({'bold': True})

            malformed_format = workbook.add_format({'bg_color': '#f4cccc'})
            well_formed_format = workbook.add_format({'bg_color': '#d9ead3'})

            disapproved_format = workbook.add_format({'bg_color': '#f4cccc'})
            approved_format = workbook.add_format({'bg_color': '#d9ead3'})
            too_short_or_long_format = workbook.add_format({'bg_color': '#fff2cc'})

            canonicalized_format = workbook.add_format({'bg_color': '#d9ead3'})

            # Sort entities by initial letters
            letters = defaultdict(lambda: list())
            for entity in entities:
                letter, *_ = str(entity)
                letter = letter.upper()
                letters[letter].append(entity)

            # Create worksheets for individual initial letters
            for letter, letter_entities in sorted(letters.items()):
                worksheet = workbook.add_worksheet(letter)

                # Set column width
                worksheet.set_column('A:A', 50)
                worksheet.set_column('B:B', 20)
                worksheet.set_column('C:C', 15)
                worksheet.set_column('D:D', 50)
                worksheet.set_column('E:E', 50)

                # Add headers
                worksheet.write('A1', 'Entita', bold_format)
                worksheet.write('B1', f'Použít? ({Approval.list_symbols()})', bold_format)
                worksheet.write_comment('B1', 'Uveďte, jestli si přejete entitu použít v '
                                        'trénovací datové sadě pro rozpoznávání jmenných '
                                        f'entit.\n\n{Approval.APPROVED}: Ano, přeji si entitu použít.\n\n'
                                        f'{Approval.DISAPPROVED}: Ne, přeji si entitu vynechat z datové sady.\n\n'
                                        f'{Approval.TOO_SHORT}: Entita je příliš krátká, pravděpodobně ji budeme '
                                        'chtít vynechat z datové sady, nebo ji budeme chtít '
                                        'přisoudit menší váhu.\n\n'
                                        f'{Approval.TOO_LONG}: Entita je příliš dlouhá, pravděpodobně ji budeme '
                                        'chtít vynechat z datové sady, nebo ji budeme chtít '
                                        'přisoudit menší váhu.')
                worksheet.write('C1', f'Typ entity ({EntityType.list_symbols()})', bold_format)
                worksheet.write_comment('C1', f'Uveďte typ entity:\n\n{EntityType.PERSON}: Osoba\n\n'
                                        f'{EntityType.PLACE}: Místo')
                worksheet.write('D1', 'Kanonická forma', bold_format)
                worksheet.write_comment('D1', 'Uveďte kanonickou formu entity, kterou '
                                        'budeme následně fulltextově hledat v dokumentech. '
                                        'Pokud lze kanonickou formu získat automatizovaně '
                                        '(např. odstraněním interpunkce zleva a zprava), '
                                        'použijte prosím poznámkový aparát.')
                worksheet.write('E1', 'Poznámky', bold_format)
                worksheet.write_comment('E1', 'Volitelně uveďte dodatečné informace ke své '
                                        'anotaci.')

                # List entities
                assert cls.MINIMUM_LENGTH <= cls.MAXIMUM_LENGTH
                for row_number, entity in enumerate(letter_entities):
                    row_number += 1
                    regest = sorted(entities_to_regests[entity])[0]

                    worksheet.write_url(row_number, 0, regest.document_url, string=str(entity))

                    canonical_form = map_entity(str(entity), entity_type)
                    if canonical_form != str(entity):
                        worksheet.write(row_number, 3, canonical_form)

                    approval = filter_entity(canonical_form, entity_type)
                    if approval != Approval.APPROVED:
                        worksheet.write(row_number, 1, str(approval))
                    else:
                        approved.add(canonical_form)

                    if canonical_form in CANONICAL_PLACES:
                        if entity_type != 'places':
                            worksheet.write(row_number, 2, 'M')
                    elif canonical_form in CANONICAL_PERSONS:
                        if entity_type != 'persons':
                            worksheet.write(row_number, 2, 'O')
                    else:
                        if approval == Approval.APPROVED:
                            worksheet.write(row_number, 1, 'N')
                            approved.remove(canonical_form)

                # Add an autofilter, so that users can reorder the entities as they need
                worksheet.autofilter(0, 0, len(letter_entities), 4)

                # Set conditional formatting for approval
                blank_conditional_format = {'type': 'blanks', 'stop_if_true': True, 'format': default_format}
                disapproved_conditional_format = {'type': 'cell', 'criteria': '==',
                                                  'value': f'"{Approval.DISAPPROVED}"',
                                                  'format': disapproved_format}
                approved_conditional_format = {'type': 'cell', 'criteria': '==',
                                               'value': f'"{Approval.APPROVED}"',
                                               'format': approved_format}
                too_short_conditional_format = {'type': 'cell', 'criteria': '==',
                                                'value': f'"{Approval.TOO_SHORT}"',
                                                'format': too_short_or_long_format}
                too_long_conditional_format = {'type': 'cell', 'criteria': '==',
                                               'value': f'"{Approval.TOO_LONG}"',
                                               'format': too_short_or_long_format}

                approval_format_range = [1, 1, len(letter_entities), 1]

                worksheet.conditional_format(*approval_format_range, blank_conditional_format)
                worksheet.conditional_format(*approval_format_range, disapproved_conditional_format)
                worksheet.conditional_format(*approval_format_range, approved_conditional_format)
                worksheet.conditional_format(*approval_format_range, too_short_conditional_format)
                worksheet.conditional_format(*approval_format_range, too_long_conditional_format)

                # Set conditional formatting for entity types
                place_conditional_format = {'type': 'cell', 'criteria': '==',
                                            'value': f'"{EntityType.PLACE}"',
                                            'format': well_formed_format}
                person_conditional_format = {'type': 'cell', 'criteria': '==',
                                             'value': f'"{EntityType.PERSON}"',
                                             'format': well_formed_format}
                malformed_conditional_format = {
                    'type': 'formula',
                    'criteria': f'=AND($C2<>"{EntityType.PERSON}", $C2<>"{EntityType.PLACE}")',
                    'stop_if_true': True, 'format': malformed_format,
                }

                correct_entity_type_format_range = [1, 2, len(letter_entities), 2]

                worksheet.conditional_format(*correct_entity_type_format_range, blank_conditional_format)
                worksheet.conditional_format(*correct_entity_type_format_range, place_conditional_format)
                worksheet.conditional_format(*correct_entity_type_format_range, person_conditional_format)
                worksheet.conditional_format(*correct_entity_type_format_range, malformed_conditional_format)

                # Set conditional formatting for canonical form
                canonicalized_conditional_format = {'type': 'no_blanks', 'format': canonicalized_format}

                canonical_form_format_range = [1, 3, len(letter_entities), 3]

                worksheet.conditional_format(*canonical_form_format_range, blank_conditional_format)
                worksheet.conditional_format(*canonical_form_format_range, canonicalized_conditional_format)

        print(len(approved))


Entity = str
Regex = str


def starfilter(fn, iterable):
    return filter(lambda x: fn(*x), iterable)


def text_to_regex(text: str, separator: str = r'\s*[,/]\s*') -> Regex:
    items = (r'\b'+escape(item)+(r'\b' if item[-1].isalpha() else '') for item in split(separator, text))
    regex = '|'.join(sorted(items, key=len, reverse=True))
    regex = f'({regex})'
    return regex


def regexes_to_regex(regexes: str, separator: str = r'\s*\n\s*') -> Regex:
    items = split(separator, regexes)
    items = (sub(r"\s+", r"\s+", item) for item in items)
    items = (sub(r"(\([^()]*\))\?\\s\+", r"(?:\1\s+)?", item) for item in items)
    items = (sub(r"\\s\+(\([^()]*\))\?", r"(?:\s+\1)?", item) for item in items)
    items = (sub(r"\.", r"\\.", item) for item in items)
    regex = '|'.join(f'({item})' for item in items)
    regex = f'({regex})'
    return regex


CLUTTER = r'[\p{Punctuation}\p{Separator}\p{Other}]*'


PREFIXES = """
auf, aus, aus der, ausz, bei, bey, bey dem, bey der, by,
civitatem/civitatis/civitati/civitate/civium, czu/czum/czur,
d., das, de, dem, den, der, des, die, di, do, dye, dy, dicti,
famosi/famoso/famosorum/famos./famosus/famosum,
für, fur, gein, gen, gen dem, gen der,
honesta/honeste/honesti/honestorum,
honorab./honorabil./honorabilem/honorab. d./honorabil. d.,
in, im, in dem, in der, in di, in die, na,
nobil, nobilem, nobilis, nobil., nobil. d., nobil d., nobil. dd.,
nobil. dom, nobilem d., nobili, nobilis, nobilis d., nobilis. d.,
nobilis dominus, nobilibus, nobilium, nobili, nobilum,
quondam, strennui, strenuum, strenuo, strenuus,
umb, urozeného, urozené, urozený, urozenému, u, v,
vf der, vf dem, vmb, von, von dem, von der, vor, vor dem,
vor das, vor den, vor die, zcu, z, ze, zu, zu dem, zu der,
zur, zum
""".strip()
PREFIXES = text_to_regex(PREFIXES)

ELLIPSIS = r'(\.{3,}|…)'

QUESTION_MARK = r'\?'

ALLOWED_SHORT_ENTITIES: Dict[str, Regex] = {
    'places': text_to_regex('Řím, Týn, Ulm'),
    'persons': text_to_regex('Iva, Jan, Lev, Ota/Otu/Oty, Vít'),
}

ROMAN_NUMBER = r'\b[MDCLXVI]+\b'
NONEXISTANT_CHAR = '¤'
OTHER_NUMBER = r'\p{Number}'
NUMBER = f'({OTHER_NUMBER}|{ROMAN_NUMBER})'
LOWERCASE_LETTER = r'\p{Lowercase_Letter}'
UPPERCASE_LETTER = r'\p{Uppercase_Letter}'
SAINT = r'sv?\.'
INCORRECT_PREFIX = f'^({LOWERCASE_LETTER}|{NUMBER})'
CORRECT_PREFIX = r"""
(abatyše)
(\w+)? (biskup|biskupa|biskupem)
(bratr|bratra|bratru|bratry|bratři)
(bruder)
(\w+)? (klášter|kláštera) (v|na)?
(closter|conuentus|conventum|conventus|conventui)
(curia|curiam) (in )?
(císař|císaře|císařem)
(castri|castro|castrum|catro)
(český|českého|českým)? (král|krále|králem|králi)
(klaster|klastr|klostr)
(herzog|herczok|herczog|herczoge|herczogen|herczogens|herczogs|herczugen|herczug|hertzog|hertzoge|hertzogen|herzog|herzoge|herzogen)
(her|herr|hern|herre|herren|herrn)
(město|města)
(dům|domu)
(\w+)? (kostel|kostela|kostele|kostelem|kostelu) (v|na)?
(frater|fratres|fratrem|fratri|fratris|fri|fri.|fris|fris.)
(graf|graff)
(domina|domine|domini|domino|dominum)
(hrad|hradě|hradu) (v|ve|na)?
(regno|regnum)
(villa|villam|villis)
(kaple|kapli) (v|na)
(kněz|kněze|knězi|knězem|kněžnou|knížete|knížectví)
(\w+)? (kníže|knížete|knížetem)
(\w+)? (kanovník|kanovníka|kanovníkem)
(marggraf|marggrafe|marggraue|marggrauen|marggraven)
(meister|mistr|mistra|mistrem|mistru)
(magister|magistri|magistro|magistrum)
(mon.|monst.|montis|monasterii|monasteriorum|monasterie|monasteria|monasterium) (in)?
(\w+)? (opat|opata|opatu)
(otec|otce)
(opido|opidum)
(pan|pán|pana|páni|panem|paní|panu|pány|pánům) (z|ze|na)?
(papež|papeže|papežem|papeži)
(převor|převora|převorem|převoru)
(regni|regem|refni|rege|regis|regnum)
(rektor|rektora|rektorem)
(vévoda|vévody|vévodovi)
""".strip()
CORRECT_PREFIX = regexes_to_regex(CORRECT_PREFIX)
CORRECT_PREFIX = f'^{CORRECT_PREFIX}\\s+({SAINT}|{UPPERCASE_LETTER})'


def map_entity(entity: Entity, entity_type: str) -> Entity:
    # 0. zachovat tečku za římským číslem
    assert '@' not in entity
    entity = sub(f'({ROMAN_NUMBER})\\.', fr'\1{NONEXISTANT_CHAR}', entity)

    # 1. odmazat interpunkci a prázdné znaky ze začátku a konce entity
    entity = sub(f'(^{CLUTTER}|{CLUTTER}$)', '', entity)

    # 2. smazat ze začátku entity předložky apod, mohou začínat malým
    # i velkým písmenem
    entity = sub(f'^{PREFIXES}{CLUTTER}', '', entity, flags=IGNORECASE)

    # 3. pokud entita obsahuje "," nebo ";"
    # smazat část od této interpunkce dál
    entity = sub(f'{CLUTTER}[,;].*$', '', entity)

    # 4. pokud entita obsahuje "..." (a více), "?", nebo "…"
    # smazat tuto interpunkci
    entity = sub(f'{CLUTTER}({ELLIPSIS}|{QUESTION_MARK}){CLUTTER}', ' ', entity)

    # ad 0. vrátit tečku za římským číslem
    entity = sub(f'({ROMAN_NUMBER}){NONEXISTANT_CHAR}', r'\1.', entity)

    return entity


def filter_entity(entity: Entity, entity_type: str) -> Approval:
    # 6. nepoužít entity, které mají 3 a méně znaků kromě [...]
    allowed_short_entities = ALLOWED_SHORT_ENTITIES[entity_type]
    if len(entity) <= 3 and not fullmatch(allowed_short_entities, entity):
        return Approval.TOO_SHORT

    # 7. pokud potom entita začíná na slovo s počátečním malým písmenem,
    # číslici nebo na I, II, III, IIII, tak entitu nepoužít, kromě začátků, za
    # kterými musí následovat velké písmeno nebo s./sv.: [...]
    if match(INCORRECT_PREFIX, entity) and not match(CORRECT_PREFIX, entity):
        return Approval.DISAPPROVED

    return Approval.APPROVED
